import numpy as np
import pandas as pd
from sklearn.decomposition import PCA, KernelPCA
from sklearn.manifold import TSNE
from sklearn.preprocessing import MinMaxScaler
from scipy.spatial import ConvexHull
from matplotlib import pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import umap
import inspect


class ImageTransformer:
    """Transform features to an image matrix using dimensionality reduction

    This class takes in data normalized between 0 and 1 and converts it to a
    CNN compatible 'image' matrix

    """

    def __init__(
        self,
        pixels,
        feature_extractor="umap",
        random_state=None,
        n_jobs=None,
        n_channels=3,
    ):
        """Generate an ImageTransformer instance

        Args:
            feature_extractor: string of value ('tsne', 'pca', 'kpca') or a
                class instance with method `fit_transform` that returns a
                2-dimensional array of extracted features.
            pixels: int (square matrix) or tuple of ints (height, width) that
                defines the size of the image matrix.
            random_state: int or RandomState. Determines the random number
                generator, if present, of a string defined feature_extractor.
            n_jobs: The number of parallel jobs to run for a string defined
                feature_extractor.
        """
        self.n_channels = n_channels
        # For each channel, we need two axis, so that we need 2*n_channels n_components to get n_channels.
        self.n_components = 2 * self.n_channels
        self.random_state = random_state
        self.n_jobs = n_jobs

        if isinstance(feature_extractor, str):
            fe = feature_extractor.casefold()
            if fe == "tsne".casefold():
                fe = TSNE(
                    n_components=self.n_components,
                    metric="cosine",
                    random_state=self.random_state,
                    n_jobs=self.n_jobs,
                    method="exact",
                )
            elif fe == "pca".casefold():
                fe = PCA(n_components=self.n_components, random_state=self.random_state)
            elif fe == "kpca".casefold():
                fe = KernelPCA(
                    n_components=self.nb_components,
                    kernel="rbf",
                    random_state=self.random_state,
                    n_jobs=self.n_jobs,
                )
            elif fe == "umap".casefold():
                fe = umap.UMAP(
                    n_components=self.n_components, random_state=self.random_state
                )

            else:
                raise ValueError(
                    ("Feature extraction method '{}' not accepted").format(
                        feature_extractor
                    )
                )
            self._fe = fe
        elif hasattr(feature_extractor, "fit_transform") and inspect.ismethod(
            feature_extractor.fit_transform
        ):
            self._fe = feature_extractor
        else:
            raise TypeError(
                "Parameter feature_extractor is not a "
                'string nor has method "fit_transform"'
            )

        if isinstance(pixels, int):
            pixels = (pixels, pixels)
        self._pixels = pixels
        self._xrot = None
        self._coords_l = [0] * self.n_channels

    def fit(self, X, y=None, plot=False):
        """Train the image transformer from the training set (X)

        Args:
            X: {array-like, sparse matrix} of shape (n_samples, n_features)
            y: Ignored. Present for continuity with scikit-learn
            plot: boolean of whether to produce a scatter plot showing the
                feature reduction, hull points, and minimum bounding rectangle

        Returns:
            self: object
        """
        # perform dimensionality reduction
        x_new_3D = self._fe.fit_transform(X.T)

        x_new_l = [x_new_3D[:, (2 * k) : (2 * (k + 1))] for k in range(self.n_channels)]

        # get the convex hull for the points for each channel
        self.x_rot_l = []
        for x_new in x_new_l:
            chvertices = ConvexHull(x_new).vertices
            hull_points = x_new[chvertices]
            # determine the minimum bounding rectangle
            mbr, mbr_rot = self._minimum_bounding_rectangle(hull_points)
            # rotate the matrix
            # save the rotated matrix in case user wants to change the pixel size
            self.x_rot_l.append(np.dot(mbr_rot, x_new.T).T)
            # determine feature coordinates based on pixel dimension
            if plot is True:
                self.plot_cvx_hull(x_new, chvertices, mbr)
        self._calculate_coords_bis()
        return self

    def plot_cvx_hull(self, x, chvertices, mbr):
        plt.scatter(
            x[:, 0],
            x[:, 1],
            s=1,
            cmap=plt.cm.get_cmap("jet", 10),
            alpha=0.2,
        )
        plt.fill(
            x[chvertices, 0],
            x[chvertices, 1],
            edgecolor="r",
            fill=False,
        )
        plt.fill(mbr[:, 0], mbr[:, 1], edgecolor="g", fill=False)
        plt.gca().set_aspect("equal", adjustable="box")
        plt.show()

    @property
    def pixels(self):
        """The image matrix dimensions

        Returns:
            tuple: the image matrix dimensions (height, width)

        """
        return self._pixels

    @pixels.setter
    def pixels(self, pixels):
        """Set the image matrix dimension

        Args:
            pixels: int or tuple with the dimensions (height, width)
            of the image matrix

        """
        if isinstance(pixels, int):
            pixels = (pixels, pixels)
        self._pixels = pixels
        # recalculate coordinates if already fit
        if hasattr(self, "_coords_l"):
            self._calculate_coords_bis()

    def _calculate_coords(self):
        """Calculate the matrix coordinates of each feature based on the
        pixel dimensions.
        """
        ax0_coord = (
            np.digitize(
                self._xrot[:, 0],
                bins=np.linspace(
                    min(self._xrot[:, 0]), max(self._xrot[:, 0]), self._pixels[0]
                ),
            )
            - 1
        )
        ax1_coord = (
            np.digitize(
                self._xrot[:, 1],
                bins=np.linspace(
                    min(self._xrot[:, 1]), max(self._xrot[:, 1]), self._pixels[1]
                ),
            )
            - 1
        )
        self._coords = np.stack((ax0_coord, ax1_coord), axis=1)

    def _calculate_coords_bis(self):
        """Calculate the matrix coordinates of each feature based on the
        pixel dimensions.
        """
        self._coords_l = []
        for x_rot in self.x_rot_l:
            ax0_coord = (
                np.digitize(
                    x_rot[:, 0],
                    bins=np.linspace(
                        min(x_rot[:, 0]), max(x_rot[:, 0]), self._pixels[0]
                    ),
                )
                - 1
            )
            ax1_coord = (
                np.digitize(
                    x_rot[:, 1],
                    bins=np.linspace(
                        min(x_rot[:, 1]), max(x_rot[:, 1]), self._pixels[1]
                    ),
                )
                - 1
            )
            self._coords_l.append(np.stack((ax0_coord, ax1_coord), axis=1))

    def transform(self, X, format="rgb", empty_value=0, pixels=None):
        """Transform the input matrix into image matrices

        Args:
            X: {array-like, sparse matrix} of shape (n_samples, n_features)
                where n_features matches the training set.
            format: The format of the image matrix to return. 'scalar' return a
                array of shape (M, N). 'rgb' returns an numpy.ndarray of shape
                (M, N, 3) that is compatible with PIL.
            empty_value: numeric value to fill elements where no features are
                mapped. Default = 0.
            pixels: int or tuple, default = None
                Will transform with as many pixels.
        Returns:
            A list of n_samples numpy matrices of dimensions set by
            the pixel parameter. I added one dimension that controls the number of channels.
        """

        img_matrices_l = list()
        if pixels is not None:
            self.pixels = pixels
        for coords in self._coords_l:
            img_coords = (
                pd.DataFrame(np.vstack((coords.T, X)).T)
                .groupby([0, 1], as_index=False)
                .mean()
            )

            img_matrices = []
            blank_mat = np.zeros(self._pixels)
            if empty_value != 0:
                blank_mat[:] = empty_value
            for z in range(2, img_coords.shape[1]):
                img_matrix = blank_mat.copy()
                img_matrix[
                    img_coords[0].astype(int), img_coords[1].astype(int)
                ] = img_coords[z]
                img_matrices.append(img_matrix)

            if format == "rgb":
                img_matrices = np.array([self._mat_to_rgb(m) for m in img_matrices])
            elif format == "scalar":
                img_matrices = np.stack(img_matrices)
            else:
                raise ValueError(
                    ("'{}' not accepted for parameter 'format'").format(format)
                )
            img_matrices_l.append(img_matrices)
        return np.stack(img_matrices_l, axis=1)

    def fit_transform(self, X, **kwargs):
        """Train the image transformer from the training set (X) and return
        the transformed data.

        Args:
            X: {array-like, sparse matrix} of shape (n_samples, n_features)

        Returns:
            A list of n_samples numpy matrices of dimensions set by
            the pixel parameter
        """
        needing_plot = kwargs.pop("plot")
        self.fit(X, plot=needing_plot)
        return self.transform(X, **kwargs)

    def feature_density_matrix(self, n_channel):
        """Generate image matrix with feature counts per pixel

        Returns:
            img_matrix (ndarray): matrix with feature counts per pixel
        """
        fdmat = np.zeros(self._pixels)
        np.add.at(fdmat, tuple(self._coords_l[n_channel].T), 1)
        return fdmat

    def show_fdm(self, n_channel):
        fdm = self.feature_density_matrix(n_channel)
        fdm[fdm == 0] = np.nan
        ax = sns.heatmap(
            fdm, cmap="viridis", linewidths=0.01, linecolor="lightgrey", square=True
        )
        ax.xaxis.set_major_locator(ticker.MultipleLocator(5))
        ax.yaxis.set_major_locator(ticker.MultipleLocator(5))
        for _, spine in ax.spines.items():
            spine.set_visible(True)
            _ = plt.title("Feature per pixel")

    def coords(self):
        """Get feature coordinates

        Returns:
            ndarray: the pixel coordinates for features
        """
        return self._coords.copy()

    @staticmethod
    def _minimum_bounding_rectangle(hull_points):
        """Find the smallest bounding rectangle for a set of points.

        Modified from JesseBuesking at https://stackoverflow.com/a/33619018
        Returns a set of points representing the corners of the bounding box.

        Args:
            hull_points : an nx2 matrix of hull coordinates

        Returns:
            (tuple): tuple containing
                coords (ndarray): coordinates of the corners of the rectangle
                rotmat (ndarray): rotation matrix to align edges of rectangle
                    to x and y
        """

        pi2 = np.pi / 2
        # calculate edge angles
        edges = hull_points[1:] - hull_points[:-1]
        angles = np.arctan2(edges[:, 1], edges[:, 0])
        angles = np.abs(np.mod(angles, pi2))
        angles = np.unique(angles)
        # find rotation matrices
        rotations = np.vstack(
            [np.cos(angles), -np.sin(angles), np.sin(angles), np.cos(angles)]
        ).T
        rotations = rotations.reshape((-1, 2, 2))
        # apply rotations to the hull
        rot_points = np.dot(rotations, hull_points.T)
        # find the bounding points
        min_x = np.nanmin(rot_points[:, 0], axis=1)
        max_x = np.nanmax(rot_points[:, 0], axis=1)
        min_y = np.nanmin(rot_points[:, 1], axis=1)
        max_y = np.nanmax(rot_points[:, 1], axis=1)
        # find the box with the best area
        areas = (max_x - min_x) * (max_y - min_y)
        best_idx = np.argmin(areas)
        # return the best box
        x1 = max_x[best_idx]
        x2 = min_x[best_idx]
        y1 = max_y[best_idx]
        y2 = min_y[best_idx]
        rotmat = rotations[best_idx]
        # generate coordinates
        coords = np.zeros((4, 2))
        coords[0] = np.dot([x1, y2], rotmat)
        coords[1] = np.dot([x2, y2], rotmat)
        coords[2] = np.dot([x2, y1], rotmat)
        coords[3] = np.dot([x1, y1], rotmat)

        return coords, rotmat

    @staticmethod
    def _mat_to_rgb(mat):
        """Convert image matrix to numpy rgb format

        Args:
            mat: {array-like} (M, N)

        Returns:
            An numpy.ndarry (M, N, 3) with orignal values repeated across
            RGB channels.
        """
        return np.repeat(mat[:, :, np.newaxis], 3, axis=2)
